/* 
   The code in this file relies heavily on compiler optimization.  The
   code was written on an x86 with GCC 2.95 19990728, and using -O2 it
   produced assembly code comparable to hand-crafted assembly. Without
   optimiation the generated code was truly awful.

   The goal is to produce an optimized flat loop without any kind of
   conditional execution. All functions are merged and all constant
   expressions are evaluated during compile-time.

 */


#include <algorithm>

typedef unsigned char  byte;
typedef byte*          bytep;
typedef byte           uint8;
typedef unsigned short uint16;
typedef unsigned int   uint32;

template<class sampleType,
    int leastSignificantBit, int numberOfSignificantBits>
struct sample {
    typedef sampleType valueType;
    
    valueType value;
    static const int lowBit = leastSignificantBit;
    static const int bitWidth = numberOfSignificantBits;
};



template<class outSampleT, class inSampleT>
inline typename outSampleT::valueType
    convertSample(typename inSampleT::valueType v) {
    typedef typename inSampleT::valueType inT;
    typedef typename outSampleT::valueType outT;

    // inT:   ??????????xxxxxxx??????
    // outT:  000000xxxxxxx0000000000

    const int validWidth = (inSampleT::bitWidth < outSampleT::bitWidth) ?
	inSampleT::bitWidth : outSampleT::bitWidth;
    
    int inTWidth = sizeof(inT)*8;

    int leftDiscardShift = inTWidth - inSampleT::bitWidth - inSampleT::lowBit;
    int rightDiscardShift = inTWidth - validWidth;
    int finalLeftShift =
	outSampleT::bitWidth + outSampleT::lowBit - validWidth;

    //cout << "leftDiscardShift=" << leftDiscardShift << endl;   // DEBUG
    //cout << "rightDiscardShift=" << rightDiscardShift << endl; // DEBUG
    //cout << "finalLeftShift=" << finalLeftShift << endl;       // DEBUG
    
    v <<= leftDiscardShift;
    v >>= rightDiscardShift;

    outT stripped = v;
    return stripped << finalLeftShift;
}



class ComponentSampleModel {};

class ComponentSampleModelByteABC : public ComponentSampleModel {
    public:
    typedef byte componentType;

    typedef sample<byte, 0, 8> aType;
    typedef sample<byte, 0, 8> bType;
    typedef sample<byte, 0, 8> cType;

    static const int o0 = 0;
    static const int o1 = 1;
    static const int o2 = 2;
};

class ComponentSampleModelByteCBA : public ComponentSampleModel {
    public:
    typedef byte componentType;

    typedef sample<byte, 0, 8> aType;
    typedef sample<byte, 0, 8> bType;
    typedef sample<byte, 0, 8> cType;

    static const int o0 = 2;
    static const int o1 = 1;
    static const int o2 = 0;
};

class SinglePixelPackedSampleModel {};

class SinglePixelPackedSampleModelUInt16_5A6B5C 
    : public SinglePixelPackedSampleModel {

    public:
    typedef uint16 unitType;
    
    typedef sample<unitType, 11, 5> aType;
    typedef sample<unitType,  5, 6> bType;
    typedef sample<unitType,  0, 5> cType;
};


template<int counter, class CSM>
    inline void consumeABC(bytep& in,
			   typename CSM::aType::valueType& a,
			   typename CSM::bType::valueType& b,
			   typename CSM::cType::valueType& c,
			   ComponentSampleModel) {
    typedef typename CSM::componentType componentType;
    componentType* componentIn = (componentType*) in;

    if (CSM::o0 == 0) a = *(componentIn++);
    if (CSM::o0 == 1) b = *(componentIn++);
    if (CSM::o0 == 2) c = *(componentIn++);

    if (CSM::o1 == 0) a = *(componentIn++);
    if (CSM::o1 == 1) b = *(componentIn++);
    if (CSM::o1 == 2) c = *(componentIn++);

    if (CSM::o2 == 0) a = *(componentIn++);
    if (CSM::o2 == 1) b = *(componentIn++);
    if (CSM::o2 == 2) c = *(componentIn++);

    in = (bytep) componentIn;
}

template<int counter, class SPPSM, class inSM>
    inline void emitABC(bytep& out,
			typename inSM::aType::valueType a,
			typename inSM::bType::valueType b,
			typename inSM::cType::valueType c,
			SinglePixelPackedSampleModel) {
    typedef typename SPPSM::unitType unitType;

    unitType* unitOut = (unitType*) out;
    
    unitType pixel =
	convertSample<typename SPPSM::aType, typename inSM::aType>(a) |
	convertSample<typename SPPSM::bType, typename inSM::bType>(b) |
	convertSample<typename SPPSM::cType, typename inSM::cType>(c);


    *(unitOut++) = pixel;

    out = (bytep) unitOut;
}

template<int counter, class inSM, class outSM>
    inline void spoolABC(bytep& in, bytep& out) {
    typename inSM::aType::valueType a;
    typename inSM::bType::valueType b;
    typename inSM::cType::valueType c;
    consumeABC<counter, inSM>(in, a, b, c, inSM());
    emitABC<counter, outSM, inSM>(out, a, b, c, outSM());
}


template<class inSM, class outSM>
    inline void renderSweeps(byte* in, byte* out,
			     int sweeps, int sweepSize,
			     int srcSweepStride, int destSweepStride) {

    for (int s=0; s<sweeps; s++) {
	byte* inS = in;
	byte* outS = out;
	for (int i=0; i<sweepSize; i++) {
	    spoolABC<0, inSM, outSM>(inS, outS);
	}
	in += srcSweepStride;
	out += destSweepStride;
    }
    
}
