
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <cni.h>
#include "no/rwr/x/Visual.h"
#include "no/rwr/x/Screen.h"
#include "no/rwr/x/Display.h"
#include <gnu/gcj/RawData.h>

using namespace gnu::gcj;

void no::rwr::x::Visual::init(RawData* visual, jint depth) {
    int givenMask = 
	((visual == 0) ? 0 : (MASK_ID | MASK_VISUAL_STRUCTURE)) |
	((depth  == 0) ? 0 : MASK_DEPTH);
    if (givenMask == 0) return;

    ensureXVisualInfo(givenMask);
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    
    if (visual != 0) {
	::Visual* visualStructure = (::Visual*) visual;
	info->visual = visualStructure;
	info->visualid = XVisualIDFromVisual(visualStructure);
    }
    if (depth  != 0) info->depth = depth;

    infoMask = givenMask;
}


void no::rwr::x::Visual::finalize() {
    if (xVisualInfo != 0) {
	delete xVisualInfo;
	xVisualInfo = 0;
    }
}


RawData* no::rwr::x::Visual::getVisualStructure() {
    ensureXVisualInfo(MASK_ALL); // Make sure structure is set
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    return reinterpret_cast<gnu::gcj::RawData*>(info->visual);
}

jint no::rwr::x::Visual::getRedMask() {
    ensureXVisualInfo(MASK_RED);
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    return info->red_mask;
}
jint no::rwr::x::Visual::getGreenMask() {
    ensureXVisualInfo(MASK_GREEN);
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    return info->green_mask;
}
jint no::rwr::x::Visual::getBlueMask() {
    ensureXVisualInfo(MASK_BLUE);
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    return info->blue_mask;
}

jint no::rwr::x::Visual::getScreenNumber() {
    if (screen != 0) {
	return screen->getScreenNumber();
    }

    ensureXVisualInfo(MASK_SCREEN);
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    return info->screen;
}

jint no::rwr::x::Visual::getDepth() {
    ensureXVisualInfo(MASK_DEPTH);
    XVisualInfo* info = (XVisualInfo*) xVisualInfo;
    return info->depth;
}



jint no::rwr::x::Visual::getVisualClass() {
    ensureXVisualInfo(MASK_CLASS);
    ::XVisualInfo* info = (::XVisualInfo*) xVisualInfo;
    return info->c_class;
}

void no::rwr::x::Visual::ensureXVisualInfo(jint requiredMask) {

    // Create structure if necessary
    if (xVisualInfo == 0) {
	XVisualInfo* info = new XVisualInfo;
	infoMask = 0;
	xVisualInfo = reinterpret_cast<gnu::gcj::RawData*>(info);
    }
    
    int missingInformation = ~infoMask;
    if ((missingInformation & requiredMask) == 0) return;

    // We need more info

    XVisualInfo* info = (XVisualInfo*) xVisualInfo;

    // Store everything we know into template
    if (screen != 0) {
	info->screen = screen->getScreenNumber();
	infoMask |= MASK_SCREEN;
    }

    // Aquire info using the current info as template for matching
    ::Display* dpy = (::Display*) display->display;
    int visualInfoCount;
    XVisualInfo* matches = XGetVisualInfo(dpy, (infoMask & MASK_ALL),
					  info, &visualInfoCount);
    if (matches != 0) {
	(*info) = matches[0];
	xVisualInfo = reinterpret_cast<gnu::gcj::RawData*>(info);
	infoMask = ~0; // ALL
	XFree(matches);
    } else {
	// FIXME: We are in trouble now. Throw exception?
    }
}
